@@ -1,7 +1,3 @@
-/* This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-
 /*! Render Bundles
 
     ## Software implementation
@@ -44,21 +40,24 @@ use crate::{
     },
     conv,
     device::{
-        AttachmentData, Device, DeviceError, RenderPassContext, MAX_VERTEX_BUFFERS,
+        AttachmentData, Device, DeviceError, MissingDownlevelFlags, RenderPassContext,
         SHADER_STAGE_COUNT,
     },
-    hub::{GfxBackend, GlobalIdentityHandlerFactory, Hub, Resource, Storage, Token},
+    error::{ErrorFormatter, PrettyError},
+    hub::{GlobalIdentityHandlerFactory, HalApi, Hub, Resource, Storage, Token},
     id,
     memory_init_tracker::{MemoryInitKind, MemoryInitTrackerAction},
-    resource::BufferUse,
+    pipeline::PipelineFlags,
     track::{TrackerSet, UsageConflict},
     validation::check_buffer_usage,
-    Label, LabelHelpers, LifeGuard, Stored, MAX_BIND_GROUPS,
+    Label, LabelHelpers, LifeGuard, Stored,
 };
 use arrayvec::ArrayVec;
-use std::{borrow::Cow, iter, mem, ops::Range};
+use std::{borrow::Cow, mem, ops::Range};
 use thiserror::Error;
 
+use hal::CommandEncoder as _;
+
 /// Describes a [`RenderBundleEncoder`].
 #[derive(Clone, Debug, Default, PartialEq, Eq, Hash)]
 #[cfg_attr(feature = "trace", derive(serde::Serialize))]
@@ -69,9 +68,9 @@ pub struct RenderBundleEncoderDescriptor<'a> {
     /// The formats of the color attachments that this render bundle is capable to rendering to. This
     /// must match the formats of the color attachments in the renderpass this render bundle is executed in.
     pub color_formats: Cow<'a, [wgt::TextureFormat]>,
-    /// The formats of the depth attachment that this render bundle is capable to rendering to. This
-    /// must match the formats of the depth attachments in the renderpass this render bundle is executed in.
-    pub depth_stencil_format: Option<wgt::TextureFormat>,
+    /// Information about the depth attachment that this render bundle is capable to rendering to. The format
+    /// must match the format of the depth attachments in the renderpass this render bundle is executed in.
+    pub depth_stencil: Option<wgt::RenderBundleDepthStencil>,
     /// Sample count this render bundle is capable of rendering to. This must match the pipelines and
     /// the renderpasses it is used in.
     pub sample_count: u32,
@@ -83,6 +82,7 @@ pub struct RenderBundleEncoder {
     base: BasePass<RenderCommand>,
     parent_id: id::DeviceId,
     pub(crate) context: RenderPassContext,
+    pub(crate) is_ds_read_only: bool,
 }
 
 impl RenderBundleEncoder {
@@ -98,16 +98,24 @@ impl RenderBundleEncoder {
                 attachments: AttachmentData {
                     colors: desc.color_formats.iter().cloned().collect(),
                     resolves: ArrayVec::new(),
-                    depth_stencil: desc.depth_stencil_format,
+                    depth_stencil: desc.depth_stencil.map(|ds| ds.format),
                 },
                 sample_count: {
                     let sc = desc.sample_count;
                     if sc == 0 || sc > 32 || !conv::is_power_of_two(sc) {
                         return Err(CreateRenderBundleError::InvalidSampleCount(sc));
                     }
-                    sc as u8
+                    sc
                 },
             },
+            is_ds_read_only: match desc.depth_stencil {
+                Some(ds) => {
+                    let aspects = hal::FormatAspects::from(ds.format);
+                    (!aspects.contains(hal::FormatAspects::DEPTH) || ds.depth_read_only)
+                        && (!aspects.contains(hal::FormatAspects::STENCIL) || ds.stencil_read_only)
+                }
+                None => false,
+            },
         })
     }
 
@@ -123,6 +131,7 @@ impl RenderBundleEncoder {
                 },
                 sample_count: 0,
             },
+            is_ds_read_only: false,
         }
     }
 
@@ -135,12 +144,12 @@ impl RenderBundleEncoder {
         self.parent_id
     }
 
-    pub(crate) fn finish<B: hal::Backend, G: GlobalIdentityHandlerFactory>(
+    pub(crate) fn finish<A: hal::Api, G: GlobalIdentityHandlerFactory>(
         self,
         desc: &RenderBundleDescriptor,
-        device: &Device<B>,
-        hub: &Hub<B, G>,
-        token: &mut Token<Device<B>>,
+        device: &Device<A>,
+        hub: &Hub<A, G>,
+        token: &mut Token<Device<A>>,
     ) -> Result<RenderBundle, RenderBundleError> {
         let (pipeline_layout_guard, mut token) = hub.pipeline_layouts.read(token);
         let (bind_group_guard, mut token) = hub.bind_groups.read(&mut token);
@@ -150,10 +159,12 @@ impl RenderBundleEncoder {
         let mut state = State {
             trackers: TrackerSet::new(self.parent_id.backend()),
             index: IndexState::new(),
-            vertex: (0..MAX_VERTEX_BUFFERS)
+            vertex: (0..hal::MAX_VERTEX_BUFFERS)
                 .map(|_| VertexState::new())
                 .collect(),
-            bind: (0..MAX_BIND_GROUPS).map(|_| BindState::new()).collect(),
+            bind: (0..hal::MAX_BIND_GROUPS)
+                .map(|_| BindState::new())
+                .collect(),
             push_constant_ranges: PushConstantState::new(),
             raw_dynamic_offsets: Vec::new(),
             flat_dynamic_offsets: Vec::new(),
@@ -214,8 +225,10 @@ impl RenderBundleEncoder {
                     state.set_bind_group(index, bind_group_id, bind_group.layout_id, offsets);
                     state
                         .trackers
-                        .merge_extend(&bind_group.used)
+                        .merge_extend_stateful(&bind_group.used)
                         .map_pass_err(scope)?;
+                    //Note: stateless trackers are not merged: the lifetime reference
+                    // is held to the bind group itself.
                 }
                 RenderCommand::SetPipeline(pipeline_id) => {
                     let scope = PassErrorScope::SetPipelineRender(pipeline_id);
@@ -231,10 +244,15 @@ impl RenderBundleEncoder {
 
                     self.context
                         .check_compatible(&pipeline.pass_context)
-                        .map_err(RenderCommandError::IncompatiblePipeline)
+                        .map_err(RenderCommandError::IncompatiblePipelineTargets)
                         .map_pass_err(scope)?;
 
-                    //TODO: check read-only depth
+                    if pipeline.flags.contains(PipelineFlags::WRITES_DEPTH_STENCIL)
+                        && self.is_ds_read_only
+                    {
+                        return Err(RenderCommandError::IncompatiblePipelineRods)
+                            .map_pass_err(scope);
+                    }
 
                     let layout = &pipeline_layout_guard[pipeline.layout_id.value];
                     pipeline_layout_id = Some(pipeline.layout_id.value);
@@ -260,9 +278,9 @@ impl RenderBundleEncoder {
                     let buffer = state
                         .trackers
                         .buffers
-                        .use_extend(&*buffer_guard, buffer_id, (), BufferUse::INDEX)
+                        .use_extend(&*buffer_guard, buffer_id, (), hal::BufferUses::INDEX)
                         .unwrap();
-                    check_buffer_usage(buffer.usage, wgt::BufferUsage::INDEX)
+                    check_buffer_usage(buffer.usage, wgt::BufferUsages::INDEX)
                         .map_pass_err(scope)?;
 
                     let end = match size {
@@ -287,9 +305,9 @@ impl RenderBundleEncoder {
                     let buffer = state
                         .trackers
                         .buffers
-                        .use_extend(&*buffer_guard, buffer_id, (), BufferUse::VERTEX)
+                        .use_extend(&*buffer_guard, buffer_id, (), hal::BufferUses::VERTEX)
                         .unwrap();
-                    check_buffer_usage(buffer.usage, wgt::BufferUsage::VERTEX)
+                    check_buffer_usage(buffer.usage, wgt::BufferUsages::VERTEX)
                         .map_pass_err(scope)?;
 
                     let end = match size {
@@ -405,12 +423,16 @@ impl RenderBundleEncoder {
                         indirect: true,
                         pipeline: state.pipeline.last_state,
                     };
+                    device
+                        .require_downlevel_flags(wgt::DownlevelFlags::INDIRECT_EXECUTION)
+                        .map_pass_err(scope)?;
+
                     let buffer = state
                         .trackers
                         .buffers
-                        .use_extend(&*buffer_guard, buffer_id, (), BufferUse::INDIRECT)
+                        .use_extend(&*buffer_guard, buffer_id, (), hal::BufferUses::INDIRECT)
                         .unwrap();
-                    check_buffer_usage(buffer.usage, wgt::BufferUsage::INDIRECT)
+                    check_buffer_usage(buffer.usage, wgt::BufferUsages::INDIRECT)
                         .map_pass_err(scope)?;
 
                     buffer_memory_init_actions.extend(
@@ -441,13 +463,17 @@ impl RenderBundleEncoder {
                         indirect: true,
                         pipeline: state.pipeline.last_state,
                     };
+                    device
+                        .require_downlevel_flags(wgt::DownlevelFlags::INDIRECT_EXECUTION)
+                        .map_pass_err(scope)?;
+
                     let buffer = state
                         .trackers
                         .buffers
-                        .use_extend(&*buffer_guard, buffer_id, (), BufferUse::INDIRECT)
+                        .use_extend(&*buffer_guard, buffer_id, (), hal::BufferUses::INDIRECT)
                         .map_err(|err| RenderCommandError::Buffer(buffer_id, err))
                         .map_pass_err(scope)?;
-                    check_buffer_usage(buffer.usage, wgt::BufferUsage::INDIRECT)
+                    check_buffer_usage(buffer.usage, wgt::BufferUsages::INDIRECT)
                         .map_pass_err(scope)?;
 
                     buffer_memory_init_actions.extend(
@@ -492,6 +518,7 @@ impl RenderBundleEncoder {
                 string_data: Vec::new(),
                 push_constant_data: Vec::new(),
             },
+            is_ds_read_only: self.is_ds_read_only,
             device_id: Stored {
                 value: id::Valid(self.parent_id),
                 ref_count: device.life_guard.add_ref(),
@@ -534,6 +561,17 @@ pub enum ExecutionError {
     #[error("using {0} in a render bundle is not implemented")]
     Unimplemented(&'static str),
 }
+impl PrettyError for ExecutionError {
+    fn fmt_pretty(&self, fmt: &mut ErrorFormatter) {
+        fmt.error(self);
+        match *self {
+            Self::DestroyedBuffer(id) => {
+                fmt.buffer_label(&id);
+            }
+            Self::Unimplemented(_reason) => {}
+        };
+    }
+}
 
 pub type RenderBundleDescriptor<'a> = wgt::RenderBundleDescriptor<Label<'a>>;
 
@@ -545,10 +583,11 @@ pub struct RenderBundle {
     // Normalized command stream. It can be executed verbatim,
     // without re-binding anything on the pipeline change.
     base: BasePass<RenderCommand>,
+    pub(super) is_ds_read_only: bool,
     pub(crate) device_id: Stored<id::DeviceId>,
     pub(crate) used: TrackerSet,
-    pub(crate) buffer_memory_init_actions: Vec<MemoryInitTrackerAction<id::BufferId>>,
-    pub(crate) context: RenderPassContext,
+    pub(super) buffer_memory_init_actions: Vec<MemoryInitTrackerAction<id::BufferId>>,
+    pub(super) context: RenderPassContext,
     pub(crate) life_guard: LifeGuard,
 }
 
@@ -565,23 +604,21 @@ impl RenderBundle {
     /// Note that the function isn't expected to fail, generally.
     /// All the validation has already been done by this point.
     /// The only failure condition is if some of the used buffers are destroyed.
-    pub(crate) unsafe fn execute<B: GfxBackend>(
+    pub(super) unsafe fn execute<A: HalApi>(
         &self,
-        cmd_buf: &mut B::CommandBuffer,
+        raw: &mut A::CommandEncoder,
         pipeline_layout_guard: &Storage<
-            crate::binding_model::PipelineLayout<B>,
+            crate::binding_model::PipelineLayout<A>,
             id::PipelineLayoutId,
         >,
-        bind_group_guard: &Storage<crate::binding_model::BindGroup<B>, id::BindGroupId>,
-        pipeline_guard: &Storage<crate::pipeline::RenderPipeline<B>, id::RenderPipelineId>,
-        buffer_guard: &Storage<crate::resource::Buffer<B>, id::BufferId>,
+        bind_group_guard: &Storage<crate::binding_model::BindGroup<A>, id::BindGroupId>,
+        pipeline_guard: &Storage<crate::pipeline::RenderPipeline<A>, id::RenderPipelineId>,
+        buffer_guard: &Storage<crate::resource::Buffer<A>, id::BufferId>,
     ) -> Result<(), ExecutionError> {
-        use hal::command::CommandBuffer as _;
-
         let mut offsets = self.base.dynamic_offsets.as_slice();
         let mut pipeline_layout_id = None::<id::Valid<id::PipelineLayoutId>>;
         if let Some(ref label) = self.base.label {
-            cmd_buf.begin_debug_marker(label, 0);
+            raw.begin_debug_marker(label);
         }
 
         for command in self.base.commands.iter() {
@@ -592,17 +629,17 @@ impl RenderBundle {
                     bind_group_id,
                 } => {
                     let bind_group = bind_group_guard.get(bind_group_id).unwrap();
-                    cmd_buf.bind_graphics_descriptor_sets(
+                    raw.set_bind_group(
                         &pipeline_layout_guard[pipeline_layout_id.unwrap()].raw,
-                        index as usize,
-                        iter::once(bind_group.raw.raw()),
-                        offsets.iter().take(num_dynamic_offsets as usize).cloned(),
+                        index as u32,
+                        &bind_group.raw,
+                        &offsets[..num_dynamic_offsets as usize],
                     );
                     offsets = &offsets[num_dynamic_offsets as usize..];
                 }
                 RenderCommand::SetPipeline(pipeline_id) => {
                     let pipeline = pipeline_guard.get(pipeline_id).unwrap();
-                    cmd_buf.bind_graphics_pipeline(&pipeline.raw);
+                    raw.set_render_pipeline(&pipeline.raw);
 
                     pipeline_layout_id = Some(pipeline.layout_id.value);
                 }
@@ -612,19 +649,18 @@ impl RenderBundle {
                     offset,
                     size,
                 } => {
-                    let index_type = conv::map_index_format(index_format);
-
-                    let &(ref buffer, _) = buffer_guard
+                    let buffer = buffer_guard
                         .get(buffer_id)
                         .unwrap()
                         .raw
                         .as_ref()
                         .ok_or(ExecutionError::DestroyedBuffer(buffer_id))?;
-                    let range = hal::buffer::SubRange {
+                    let bb = hal::BufferBinding {
+                        buffer,
                         offset,
-                        size: size.map(|s| s.get()),
+                        size,
                     };
-                    cmd_buf.bind_index_buffer(buffer, range, index_type);
+                    raw.set_index_buffer(bb, index_format);
                 }
                 RenderCommand::SetVertexBuffer {
                     slot,
@@ -632,17 +668,18 @@ impl RenderBundle {
                     offset,
                     size,
                 } => {
-                    let &(ref buffer, _) = buffer_guard
+                    let buffer = buffer_guard
                         .get(buffer_id)
                         .unwrap()
                         .raw
                         .as_ref()
                         .ok_or(ExecutionError::DestroyedBuffer(buffer_id))?;
-                    let range = hal::buffer::SubRange {
+                    let bb = hal::BufferBinding {
+                        buffer,
                         offset,
-                        size: size.map(|s| s.get()),
+                        size,
                     };
-                    cmd_buf.bind_vertex_buffers(slot, iter::once((buffer, range)));
+                    raw.set_vertex_buffer(slot, bb);
                 }
                 RenderCommand::SetPushConstant {
                     stages,
@@ -659,20 +696,15 @@ impl RenderBundle {
                         let data_slice = &self.base.push_constant_data
                             [(values_offset as usize)..values_end_offset];
 
-                        cmd_buf.push_graphics_constants(
-                            &pipeline_layout.raw,
-                            conv::map_shader_stage_flags(stages),
-                            offset,
-                            &data_slice,
-                        )
+                        raw.set_push_constants(&pipeline_layout.raw, stages, offset, data_slice)
                     } else {
                         super::push_constant_clear(
                             offset,
                             size_bytes,
                             |clear_offset, clear_data| {
-                                cmd_buf.push_graphics_constants(
+                                raw.set_push_constants(
                                     &pipeline_layout.raw,
-                                    conv::map_shader_stage_flags(stages),
+                                    stages,
                                     clear_offset,
                                     clear_data,
                                 );
@@ -686,10 +718,7 @@ impl RenderBundle {
                     first_vertex,
                     first_instance,
                 } => {
-                    cmd_buf.draw(
-                        first_vertex..first_vertex + vertex_count,
-                        first_instance..first_instance + instance_count,
-                    );
+                    raw.draw(first_vertex, vertex_count, first_instance, instance_count);
                 }
                 RenderCommand::DrawIndexed {
                     index_count,
@@ -698,10 +727,12 @@ impl RenderBundle {
                     base_vertex,
                     first_instance,
                 } => {
-                    cmd_buf.draw_indexed(
-                        first_index..first_index + index_count,
+                    raw.draw_indexed(
+                        first_index,
+                        index_count,
                         base_vertex,
-                        first_instance..first_instance + instance_count,
+                        first_instance,
+                        instance_count,
                     );
                 }
                 RenderCommand::MultiDrawIndirect {
@@ -710,13 +741,13 @@ impl RenderBundle {
                     count: None,
                     indexed: false,
                 } => {
-                    let &(ref buffer, _) = buffer_guard
+                    let buffer = buffer_guard
                         .get(buffer_id)
                         .unwrap()
                         .raw
                         .as_ref()
                         .ok_or(ExecutionError::DestroyedBuffer(buffer_id))?;
-                    cmd_buf.draw_indirect(buffer, offset, 1, 0);
+                    raw.draw_indirect(buffer, offset, 1);
                 }
                 RenderCommand::MultiDrawIndirect {
                     buffer_id,
@@ -724,13 +755,13 @@ impl RenderBundle {
                     count: None,
                     indexed: true,
                 } => {
-                    let &(ref buffer, _) = buffer_guard
+                    let buffer = buffer_guard
                         .get(buffer_id)
                         .unwrap()
                         .raw
                         .as_ref()
                         .ok_or(ExecutionError::DestroyedBuffer(buffer_id))?;
-                    cmd_buf.draw_indexed_indirect(buffer, offset, 1, 0);
+                    raw.draw_indexed_indirect(buffer, offset, 1);
                 }
                 RenderCommand::MultiDrawIndirect { .. }
                 | RenderCommand::MultiDrawIndirectCount { .. } => {
@@ -755,7 +786,7 @@ impl RenderBundle {
         }
 
         if let Some(_) = self.base.label {
-            cmd_buf.end_debug_marker();
+            raw.end_debug_marker();
         }
 
         Ok(())
@@ -832,7 +863,7 @@ struct VertexState {
     buffer: Option<id::BufferId>,
     range: Range<wgt::BufferAddress>,
     stride: wgt::BufferAddress,
-    rate: wgt::InputStepMode,
+    rate: wgt::VertexStepMode,
     is_dirty: bool,
 }
 
@@ -842,7 +873,7 @@ impl VertexState {
             buffer: None,
             range: 0..0,
             stride: 0,
-            rate: wgt::InputStepMode::Vertex,
+            rate: wgt::VertexStepMode::Vertex,
             is_dirty: false,
         }
     }
@@ -905,7 +936,7 @@ impl BindState {
 
 #[derive(Debug)]
 struct PushConstantState {
-    ranges: ArrayVec<[wgt::PushConstantRange; SHADER_STAGE_COUNT]>,
+    ranges: ArrayVec<wgt::PushConstantRange, { SHADER_STAGE_COUNT }>,
     is_dirty: bool,
 }
 impl PushConstantState {
@@ -943,8 +974,8 @@ struct VertexLimitState {
 struct State {
     trackers: TrackerSet,
     index: IndexState,
-    vertex: ArrayVec<[VertexState; MAX_VERTEX_BUFFERS]>,
-    bind: ArrayVec<[BindState; MAX_BIND_GROUPS]>,
+    vertex: ArrayVec<VertexState, { hal::MAX_VERTEX_BUFFERS }>,
+    bind: ArrayVec<BindState, { hal::MAX_BIND_GROUPS }>,
     push_constant_ranges: PushConstantState,
     raw_dynamic_offsets: Vec<wgt::DynamicOffset>,
     flat_dynamic_offsets: Vec<wgt::DynamicOffset>,
@@ -966,13 +997,13 @@ impl State {
             }
             let limit = ((vbs.range.end - vbs.range.start) / vbs.stride) as u32;
             match vbs.rate {
-                wgt::InputStepMode::Vertex => {
+                wgt::VertexStepMode::Vertex => {
                     if limit < vert_state.vertex_limit {
                         vert_state.vertex_limit = limit;
                         vert_state.vertex_limit_slot = idx as _;
                     }
                 }
-                wgt::InputStepMode::Instance => {
+                wgt::VertexStepMode::Instance => {
                     if limit < vert_state.instance_limit {
                         vert_state.instance_limit = limit;
                         vert_state.instance_limit_slot = idx as _;
@@ -1012,7 +1043,7 @@ impl State {
     fn set_pipeline(
         &mut self,
         index_format: Option<wgt::IndexFormat>,
-        vertex_strides: &[(wgt::BufferAddress, wgt::InputStepMode)],
+        vertex_strides: &[(wgt::BufferAddress, wgt::VertexStepMode)],
         layout_ids: &[id::Valid<id::BindGroupLayoutId>],
         push_constant_layouts: &[wgt::PushConstantRange],
     ) {
@@ -1114,6 +1145,8 @@ pub(super) enum RenderBundleErrorInner {
     ResourceUsageConflict(#[from] UsageConflict),
     #[error(transparent)]
     Draw(#[from] DrawError),
+    #[error(transparent)]
+    MissingDownlevelFlags(#[from] MissingDownlevelFlags),
 }
 
 impl<T> From<T> for RenderBundleErrorInner
@@ -1140,6 +1173,14 @@ impl RenderBundleError {
         inner: RenderBundleErrorInner::Device(DeviceError::Invalid),
     };
 }
+impl PrettyError for RenderBundleError {
+    fn fmt_pretty(&self, fmt: &mut ErrorFormatter) {
+        // This error is wrapper for the inner error,
+        // but the scope has useful labels
+        fmt.error(self);
+        self.scope.fmt_pretty(fmt);
+    }
+}
 
 impl<T, E> MapPassErr<T, RenderBundleError> for Result<T, E>
 where
@@ -1218,7 +1259,7 @@ pub mod bundle_ffi {
     #[no_mangle]
     pub unsafe extern "C" fn wgpu_render_bundle_set_push_constants(
         pass: &mut RenderBundleEncoder,
-        stages: wgt::ShaderStage,
+        stages: wgt::ShaderStages,
         offset: u32,
         size_bytes: u32,
         data: *const u8,
@@ -1301,7 +1342,7 @@ pub mod bundle_ffi {
     }
 
     #[no_mangle]
-    pub extern "C" fn wgpu_render_pass_bundle_indexed_indirect(
+    pub extern "C" fn wgpu_render_bundle_draw_indexed_indirect(
         bundle: &mut RenderBundleEncoder,
         buffer_id: id::BufferId,
         offset: BufferAddress,
